from typing import Optional
from datetime import datetime
from haystack_integrations.document_stores.qdrant import QdrantDocumentStore
from haystack_integrations.components.embedders.fastembed import FastembedTextEmbedder, FastembedDocumentEmbedder
from haystack_integrations.components.retrievers.qdrant import QdrantEmbeddingRetriever
from haystack.components.writers import DocumentWriter
from haystack.document_stores.types import DuplicatePolicy
from haystack.dataclasses import Document
from haystack.utils import Secret
from ..birbtalk import BirbTool

# Save and Recall facts about user
def save_load_memory_tools(qdrant_url: Optional[str] = None, 
                qdrant_key: Optional[str] = None,
                collection_name: str = "memories",
                embedding_model: str = "BAAI/bge-small-en-v1.5",
                embedding_cache_dir: Optional[str] = None,
                embedding_local_only: bool = False,
                embedding_dim: int = 384,
                n_memories: int = 10,
                chat_id: Optional[str] = None):

    # Define document store arguments
    qdrant_kwargs = {
        "index": collection_name,
        "wait_result_from_api": True,
        "embedding_dim": embedding_dim,
        "hnsw_config": {
            "payload_m": 16,
            "m": 0
        },
        "payload_fields_to_index": [{
            "field_name": "meta.chat_id",
            "field_schema": {"type": "keyword", "is_tenant": True}
        }]
    }
    qdrant_args = []
    if qdrant_url: qdrant_kwargs["url"] = qdrant_url
    else: qdrant_args.append(":memory:")
    if qdrant_key: qdrant_kwargs["api_key"] = Secret.from_token(qdrant_key)

    # Define filters
    if chat_id:
        filters = {"field": "meta.chat_id", "operator": "==", "value": chat_id}
    else: filters = None

    # Define components
    document_store = QdrantDocumentStore(*qdrant_args, **qdrant_kwargs)
    text_embedder = FastembedTextEmbedder(
        model=embedding_model,
        cache_dir=embedding_cache_dir,
        local_files_only=embedding_local_only
    )
    retriever = QdrantEmbeddingRetriever(
        document_store=document_store,
        filters=filters,
        top_k=n_memories
    )
    doc_embedder = FastembedDocumentEmbedder(
        model=embedding_model,
        cache_dir=embedding_cache_dir,
        local_files_only=embedding_local_only
    )
    writer = DocumentWriter(document_store=document_store, policy=DuplicatePolicy.NONE)
    doc_embedder.warm_up()
    text_embedder.warm_up()

    # Define load function
    def load_function(query: str, chat_id: Optional[int] = None):
        query = text_embedder.run(text=query)["embedding"]
        documents = retriever.run(query_embedding=query)["documents"]
        if documents:
            text = [f"- {x.content}" for x in documents]
            text = "\n".join(text)
            text = f"# Memories\n\n{text}"
        else:
            text = "Nothing found!"
        return text

    # Define save function
    def save_function(fact: str, chat_id: Optional[int] = None):
        documents = [Document(content=fact, meta={
            "chat_id": chat_id,
            "datetime": datetime.now().isoformat()
        })]
        documents = doc_embedder.run(documents=documents)["documents"]
        writer.run(documents=documents)
        return "Memory saved!"

    # Define tools
    return BirbTool(
        name = "save_memory",
        description = "Save short facts about user in memory.",
        function = save_function,
        arguments = {
            "fact": {
                "type": "string",
                "description": "Fact about user such as their preferences, hobbies, life conditions and etc. Should be short and informative. Only use English."
            }
        },
        required = ["fact"]
    ), BirbTool(
        name = "search_memory",
        description = "Search for facts about user such as their preferences, hobbies or life conditions in memory.",
        function = load_function, 
        arguments = {
            "query": {
                "type": "string",
                "description": "Query for performing search in vector database. Should be specific and contain as much keywords as possible. Only use English."
        }
        },
        required = ["query"]
    )



from typing import List, Optional, Literal, Callable
from dataclasses import field, dataclass
from .langgpt import langgpt

# Dataclass for tools used by llm
@dataclass
class BirbTool:
    name: str
    description: str
    function: Callable
    arguments: dict = field(default_factory={})
    required: list[str] = field(default_factory=[])

# Dataclass for system prompt template and variables
@dataclass
class BirbPrompt:
    name: str = "Birb"
    author: Optional[str] = None
    version: Optional[str] = "1.0"
    language: str = "English"
    description: str = "You are a helpful AI assistant."
    background: Optional[str] = None
    rules: Optional[List[str]] = None
    workflow: Optional[List[str]] = None
    skills: Optional[List[str]] = None
    examples: Optional[List[dict[Literal["title", "text"], str]]] = None
    data: Optional[List[dict[Literal["title", "text"], str]]] = None
    init_message: Optional[str] = "As a/an <Role>, you must follow the <Rules>, you must talk to user in default <Language>."
    template: Optional[str] = langgpt



langgpt = """# Role: {{name}}

## Profile:
  
{% if author %}- Author: {{author}}\n{% endif %}
{%- if version %}- Version: {{version}}\n{% endif %}
{%- if language %}- Language: {{language}}{% endif %}
- Description: {{description}}

{%- if background %}\n\n### Background\n\n{{background}}{% endif %}

{%- if rules %}\n\n## Rules
{% for item in rules %} 
{{loop.index}}. {{item}}
{%- endfor %}
{%- endif %}

{%- if skills %}\n\n## Skills
{% for item in skills %} 
{{loop.index}}. {{item}}
{%- endfor %}
{%- endif %}

{%- if workflow %}\n\n## Workflow
{% for item in workflow %} 
{{loop.index}}. {{item}}
{%- endfor %}
{%- endif %}

{%- if examples %}\n\n## Examples
{%- for item in examples %} 
\n### {{item.title}}\n\n{{item.text}}
{%- endfor %}
{%- endif %}

{%- if data -%}
{%- for item in data %}\n\n## {{item.title}}

{{item.text}}
{%- endfor %}
{%- endif %}

{%- if init_message %}\n\n## Initialization

{{init_message}}
{%- endif -%}
"""



from haystack.dataclasses import ChatMessage, ChatRole 
from typing import List 
from haystack import component
from jinja2 import Template
import tiktoken

# Trim message history based on max length
@component
class BufferTrimmer:
    def __init__(self, max_length: int, tiktoken_model: str = "gpt-4o") -> None:
        self.max_length = max_length
        self.encoder = tiktoken.encoding_for_model(tiktoken_model)
    @component.output_types(output=List[ChatMessage])
    def run(self, messages: List[ChatMessage]):
        # Trim
        n_tokens = 0
        for i in range(1, len(messages) + 1):
            text = messages[-i].text
            if not text: continue
            n_tokens += len(self.encoder.encode(text))
            if n_tokens > self.max_length:
                messages = messages[-i:]
                break
        # Ensure that buffer starts with user message
        start = 0
        for i, message in enumerate(messages):
            if message.is_from(ChatRole.USER):
                start = i
                break
        messages = messages[start:]
        return {"output": messages}

# Join multiple sequential user messages together
@component
class UserMessageJoiner:
    def __init__(self, multiuser: bool = False, template: str | None = None) -> None:
        self.multiuser = multiuser
        self.template = Template(template) if template else Template("{{name}}: {{text}}") 
    @component.output_types(output=List[ChatMessage])
    def run(self, messages: List[ChatMessage]):
        # Create a join mask
        mask: list[bool] = []
        for i, message in enumerate(messages):
            if i == 0 or not message.is_from(ChatRole.USER) or not message.text:
                mask.append(False)
            elif message.role == messages[i-1].role:
                if self.multiuser:
                    if messages[i-1].name == message.name:
                        mask.append(True)
                    else:
                        mask.append(False)
                else:
                    mask.append(True)
            else:
                mask.append(False)
        # Join messages
        new_messages = []
        for i, message in enumerate(messages):
            name = message.name if message.name else "User"
            if not mask[i]:
                new_message = message
                if i != len(mask) - 1 and mask[i+1]:
                    new_text = self.template.render(name=name, text=message.text)
                    new_message = ChatMessage.from_user(new_text, meta=message.meta)
                new_messages.append(new_message)
            else:
                new_text = self.template.render(name=name, text=message.text)
                new_text = f"{new_messages[-1].text}\n{new_text}"
                new_messages[-1] = ChatMessage.from_user(new_text, meta=new_messages[-1].meta)
        return {"output": new_messages}



from typing import List, Optional, Tuple
from haystack.dataclasses import ChatMessage, ChatRole
from haystack.tools import Tool
from haystack.components.generators.chat import OpenAIChatGenerator
from haystack.components.tools import ToolInvoker
from haystack.utils import Secret
from jinja2 import Template
from dataclasses import asdict
import logging
import json
from .models import BirbTool, BirbPrompt
from .components import UserMessageJoiner, BufferTrimmer
logger = logging.getLogger("BirbTalk")

# Default BirbTalk class with short-term memory buffer and tools 
class BirbTalk:
    def __init__(self,
                 api_key: Optional[str] = None,
                 api_url: Optional[str] = None,
                 model: str = "gpt-4o",
                 system_prompt: BirbPrompt | None = None,
                 buffer_size: int = 512,
                 multiuser: bool = False,
                 tools_strict: bool = True,
                 temperature: float = 1.0,
                 chat_id: Optional[str] = None,
                 is_claude: bool = False,
                 tools: List[BirbTool] = [],
                 ):
        assert api_key, "Api key is required"
        self.chat_id = chat_id
        self.buffer = []
        self.is_claude = is_claude
        self.input_tokens = 0
        self.output_tokens = 0
        # Generate system message
        system_prompt = system_prompt if system_prompt else BirbPrompt()
        self.system_prompt = Template(system_prompt.template).render(**asdict(system_prompt))
        self.system_message = ChatMessage.from_system(self.system_prompt)
        # Convert BirbTools to haystack tools
        self.tools = [Tool(
            name=x.name,
            description=x.description,
            function=x.function,
            parameters={
                "type": "object",
                "properties": x.arguments,
                "required": x.required,
                "additionalProperties": False
            }
        ) for x in tools]
        # Initialize components
        self.message_joiner = UserMessageJoiner(multiuser=multiuser)
        self.generator = OpenAIChatGenerator(
            api_key = Secret.from_token(api_key),
            api_base_url = api_url, 
            model = model, 
            generation_kwargs = {
                "temperature": temperature
            },
            tools_strict = tools_strict
        )
        self.invoker = ToolInvoker(tools = self.tools) if self.tools else None
        self.trimmer = BufferTrimmer(max_length=buffer_size)

    # Generate chat completion
    def generate(self, prompt: str, name: str = "User"):
        # Trim buffer
        self.buffer = self.trimmer.run(messages=self.buffer)["output"]
        # Add user message to buffer
        self.buffer.append(ChatMessage.from_user(prompt, name=name))
        # Construct and pre-process message list
        messages = [self.system_message] + self.buffer
        messages = self.message_joiner.run(messages)["output"]
        # Generate first reply
        resp = self.generator.run(messages=messages, tools=self.tools)
        messages += resp["replies"]
        # Count tokens
        usage = resp["meta"]["usage"]
        self.input_tokens += usage["prompt_tokens"]
        self.output_tokens += usage["completion_tokens"]
        # Process tool calls
        if len(messages[-1].tool_calls) > 0 and self.invoker:
            # Add chat id to tool call
            if self.chat_id:
                for i in range(len(messages[-1].tool_calls)):
                    messages[-1].tool_calls[i].arguments["chat_id"] = self.chat_id
            # Call tools and save results
            tool_responses = self.invoker.run(messages=messages)["tool_messages"]
            if not self.is_claude:
                messages.extend(tool_responses)
            # Claude-specific fixes
            else:
                tool_responses = "\n\n".join([x.result for x in tool_responses])
                messages = messages[:-1]
                messages[-1].text += f"\n\n---\n\n# Context:\n\n{tool_responses}"
            # Generate new reply
            resp = self.generator.run(messages=messages)
            messages += resp["replies"]
            # Count tokens
            usage = resp["meta"]["usage"]
            self.input_tokens += usage["prompt_tokens"]
            self.output_tokens += usage["completion_tokens"]
        # Add assistant reply to the buffer
        self.buffer.append(messages[-1])
        # Print debug message and return reply text
        self.debug(messages)
        return messages[-1].text
    
    # Pretty-print context
    def debug(self, messages):
        debug_text = []
        for message in messages:
            text = message.text
            if len(message.tool_calls) > 0:
                text = json.dumps(message.to_openai_dict_format())
            elif message.is_from(ChatRole.TOOL):
                text = f'{message._role.value}: {json.dumps(message.tool_call_result.result)}'
            elif not message.is_from(ChatRole.SYSTEM):
                text = f'{message._role.value}: {message.text}'
            debug_text.append(text)
        debug_text = "\n\n".join(debug_text)
        logger.debug(f"\n{debug_text}")
    
    # Dump buffer in openai format
    def dump_buffer(self) -> List[dict]:
        return [x.to_openai_dict_format() for x in self.buffer]
    
    # Load buffer from openai format
    def load_buffer(self, buffer: List[dict]):
        self.buffer = [ChatMessage.from_openai_dict_format(x) for x in buffer]

    # Get usage stats
    def get_stats(self) -> Tuple[int, int]:
        return self.input_tokens, self.output_tokens



import os
from duckduckgo_search import DDGS
from ..models import BirbTool

PROXY = os.getenv("PROXY")

# Search pages in duckduckgo
def duckduckgo_tool(max_results: int):
    # Define function
    def function(query: str, **kwargs):
        # Search
        try:
            ddgs = DDGS(proxy=PROXY.replace("socks5", "socks5h") if PROXY else None)
            results = ddgs.text(query, region='wt-wt', safesearch='off', timelimit='y', max_results=max_results)
        except:
            return "Nothing found!"
        # No results
        if len(results) == 0:
            return "Nothing found!"
        # Format texts
        texts = []
        for x in results:
            texts.append(f'Title: {x["title"]}\nSource: [{x["href"]}]({x["href"]})\nText: {x["body"]}')
        return "\n\n".join(texts)
    # Define tool
    return BirbTool(
        name = "web_search",
        description = "Search for information on the web",
        function = function,
        arguments = {
            "query": {
                "type": "string",
                "description": "Short query for search engine that reflects the user's intentions. Should be rich with keywords."
            }
        },
        required = ["query"]
    )

# Search news in duckduckgo
def duckduck_news_tool(max_results: int):
    # Define function
    def function(keywords: str, **kwargs):
        # Search
        try:
            ddgs = DDGS(proxy=PROXY.replace("socks5", "socks5h") if PROXY else None)
            results = ddgs.news(keywords, region='wt-wt', safesearch='off', timelimit='y', max_results=max_results)
        except:
            return "Nothing found!"
        # No results
        if len(results) == 0:
            return "Nothing found!"
        # Format texts
        texts = []
        for x in results:
            texts.append(f'Title: {x["title"]}\nSource: [{x["source"]}]({x["url"]})\nText: {x["body"]}')
        return "\n\n".join(texts)
    # Define tool
    return BirbTool(
        name = "get_news",
        description = "Search for news by keywords",
        function = function,
        arguments = {
            "keywords": {
                "type": "string",
                "description": "List of keywords separated by semicolons."
            }
        },
        required = ["keywords"]
    )
